using System;
using System.Collections.Generic;
using System.IO;
using Weland;

namespace GeneratorCommon
{
	public class Helpers
	{
		public static readonly byte WaterCollection = 17;
		public static readonly byte LavaCollection = 18;
		public static readonly byte SewageCollection = 19;
		public static readonly byte JjaroCollection = 20;
		public static readonly byte PfhorCollection = 21;

		public static readonly byte DayCollection = 27;
		public static readonly byte NightCollection = 28;
		public static readonly byte MoonCollection = 29;
		public static readonly byte SpaceCollection = 30;

		private static readonly Dictionary<byte, byte[]> _wallTextures;
		private static readonly Dictionary<byte, byte[]> _floorTextures;
		private static readonly Dictionary<byte, byte> _lightTexture;
		private static readonly Dictionary<byte, byte> _doorTexture;
		private static readonly Random _random;

		private readonly Level _level;

		static Helpers()
		{
			_random = new Random();

			_wallTextures = new Dictionary<byte, byte[]>();
			_wallTextures[WaterCollection] = new byte[] { 6, 11, 12, 13, 14, 17, 18, 20, 21, 22, 23, 24, 26, 28 };
			_wallTextures[LavaCollection] = new byte[] { 5, 7, 8, 11, 13, 14, 15, 16, 17, 18, 19, 20, 22, 23, 24, 25, 26, 27, 28 };
			_wallTextures[SewageCollection] = new byte[] { 5, 6, 7, 8, 10, 11, 12, 14, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28 };
			_wallTextures[JjaroCollection] = new byte[] { 5, 6, 7, 8, 10, 11, 12, 14, 15, 16, 18, 19, 20, 21, 22, 25, 26, 27, 28, 30, 31 };
			_wallTextures[PfhorCollection] = new byte[] { 6, 7, 8, 9, 11, 12, 13, 14, 16, 17, 18, 19, 20, 21, 22, 24, 25, 26, 28, 29 };

			_floorTextures = new Dictionary<byte, byte[]>();
			_floorTextures[WaterCollection] = new byte[] { 5, 6, 11, 13, 14, 17, 26, 28 };
			_floorTextures[LavaCollection] = new byte[] { 5, 7, 14, 16, 19, 20, 23, 28 };
			_floorTextures[SewageCollection] = new byte[] { 11, 14, 17, 18, 19, 23, 26 };
			_floorTextures[JjaroCollection] = new byte[] { 11, 15, 18, 19, 26 };
			_floorTextures[PfhorCollection] = new byte[] { 6, 9, 14, 18, 25, 28 };

			_lightTexture = new Dictionary<byte, byte>();
			_lightTexture[WaterCollection] = 25;
			_lightTexture[LavaCollection] = 10;
			_lightTexture[SewageCollection] = 15;
			_lightTexture[JjaroCollection] = 32;
			_lightTexture[PfhorCollection] = 10;

			_doorTexture = new Dictionary<byte, byte>();
			_doorTexture[WaterCollection] = 29;
			_doorTexture[LavaCollection] = 21;
			_doorTexture[SewageCollection] = 29;
			_doorTexture[JjaroCollection] = 29;
			_doorTexture[PfhorCollection] = 23;
		}

		public Helpers(Editor editor)
		{
			_level = editor.Level;
		}

		public static int RandInt(int minNum, int maxNum)
		{
			return _random.Next(minNum, maxNum);
		}

		public static short Rand(int minNum, int maxNum)
		{
			return (short)_random.Next(minNum, maxNum);
		}

		public static bool RandPercent(short percent)
		{
			if (Rand(0, 100) < percent)
				return true;

			return false;
		}

		public static byte GetRandomCollection()
		{
			return (byte)Rand(17, 22);
		}

		public static byte GetRandomLandscapeCollection()
		{
			return (byte)Rand(27, 31);
		}

		public static byte GetRandomWallTexture(byte collection)
		{
			var num = (byte)Rand(0, _wallTextures[collection].Length);
			return _wallTextures[collection][num];
		}

		public static byte GetRandomFloorTexture(byte collection)
		{
			var num = (byte)Rand(0, _floorTextures[collection].Length);
			return _floorTextures[collection][num];
		}

		public static byte[] GetRandomizedFloorTextures(byte collection)
		{
			var list = _floorTextures[collection];
			list.Shuffle();
			return list;
		}

		public static byte[] GetRandomizedWallTextures(byte collection)
		{
			var list = _wallTextures[collection];
			list.Shuffle();
			return list;
		}

		public static void SetFloorTexture(Polygon poly, byte collection, byte texture)
		{
			var shapeDescriptor = new ShapeDescriptor();
			shapeDescriptor.Collection = collection;
			shapeDescriptor.Bitmap = texture;
			poly.FloorTexture = shapeDescriptor;
		}

		public static void SetCeilingTexture(Polygon poly, byte collection, byte texture)
		{
			var shapeDescriptor = new ShapeDescriptor();
			shapeDescriptor.Collection = collection;
			shapeDescriptor.Bitmap = texture;
			poly.CeilingTexture = shapeDescriptor;
			poly.CeilingTransferMode = 0;
		}

		public static void SetCeilingLandscapeTexture(Polygon poly, byte collection)
		{
			var shapeDescriptor = new ShapeDescriptor();
			shapeDescriptor.Collection = collection;
			shapeDescriptor.Bitmap = 0;
			poly.CeilingTexture = shapeDescriptor;
			poly.CeilingTransferMode = 9;
		}

		public static void SetWallTexture(Side side, byte collection, byte texture)
		{
			var shapeDescriptor = new ShapeDescriptor();
			shapeDescriptor.Collection = collection;
			shapeDescriptor.Bitmap = texture;
			side.Primary.Texture = shapeDescriptor;
			side.Secondary.Texture = shapeDescriptor;
		}

		public void SetUntexturedWallTextures(byte collection, byte texture)
		{
			_level.Pave();

			var shapeDescriptor = new ShapeDescriptor();
			shapeDescriptor.Collection = collection;
			shapeDescriptor.Bitmap = texture;

			foreach (var side in _level.Sides)
			{
				if (side.Primary.Texture.Collection == WaterCollection && side.Primary.Texture.Bitmap == 5)
				{
					side.Primary.Texture = shapeDescriptor;
				}
				if (side.Secondary.Texture.Collection == WaterCollection && side.Secondary.Texture.Bitmap == 5)
				{
					side.Secondary.Texture = shapeDescriptor;
				}
			}
		}

		public void SetWallTextures(Polygon poly, byte collection, byte texture, bool includeSurroundingSides = false,
								   bool surroundingSidesOnly = false)
		{
			_level.Pave();

			var sides = GetSidesForPolygon(poly, includeSurroundingSides, surroundingSidesOnly);
			foreach (var side in sides)
			{
				SetWallTexture(side, collection, texture);
			}
		}

		public List<Side> GetSidesForPolygon(Polygon poly, bool includeSurroundingSides = false,
											bool surroundingSidesOnly = false)
		{
			var sides = new List<Side>();
			var polyIndex = GetPolygonIndex(poly);

			foreach (var lineIndex in poly.LineIndexes)
			{
				if (lineIndex >= 0)
				{
					var line = _level.Lines[lineIndex];
					if (line.ClockwisePolygonSideIndex > -1)
					{
						var side = _level.Sides[line.ClockwisePolygonSideIndex];
						if (!side.Empty() && ((includeSurroundingSides && side.PolygonIndex != polyIndex) ||
											  (side.PolygonIndex == polyIndex && !surroundingSidesOnly)))
							sides.Add(side);
					}
					if (line.CounterclockwisePolygonSideIndex > -1)
					{
						var side = _level.Sides[line.CounterclockwisePolygonSideIndex];
						if (!side.Empty() && ((includeSurroundingSides && side.PolygonIndex != polyIndex) ||
											  (side.PolygonIndex == polyIndex && !surroundingSidesOnly)))
							sides.Add(side);
					}
				}
			}
			return sides;
		}

		public void TextureTeleporters(byte collection)
		{
			var light = new Light
			{
				PrimaryActive = new Light.Function(LightingFunction.Constant, 1, 5, .8, .1),
				SecondaryActive = new Light.Function(LightingFunction.Constant, 1, 5, .7, .1),
			};
			_level.Lights.Add(light);
			var lightIndex = (short)(_level.Lights.Count - 1);

			foreach (var poly in _level.Polygons)
			{
				if (poly.Type == PolygonType.Teleporter)
				{
					SetFloorTexture(poly, collection, _lightTexture[collection]);
					poly.FloorLight = lightIndex;

					SetWallTextures(poly, collection, _lightTexture[collection]);
					var sides = GetSidesForPolygon(poly);
					foreach (var side in sides)
					{
						side.PrimaryLightsourceIndex = lightIndex;
						side.SecondaryLightsourceIndex = lightIndex;
					}

					if (HasLoweredCeiling(poly))
					{
						SetCeilingTexture(poly, collection, _lightTexture[collection]);
						poly.CeilingLight = lightIndex;
					}
				}
			}
		}

		public void TexturePlatforms(byte? collection = null)
		{
			for (short i = 0; i < _level.Polygons.Count; i++)
			{
				var poly = _level.Polygons[i];
				if (poly.Type == PolygonType.Platform)
				{
					Platform platform = null;
					foreach (var p in _level.Platforms)
					{
						if (p.PolygonIndex == i)
						{
							platform = p;
							break;
						}
					}

					if (platform != null && !platform.IsSecret)
					{
						SetFloorTexture(poly, 18, 6);
						SetWallTextures(poly, 18, 6, true);

						// Attempt to set door textures
						if (collection.HasValue && platform.IsDoor)
						{
							SetCeilingTexture(poly, 18, 6);
							SetWallTextures(poly, collection.Value, _doorTexture[collection.Value], true, true);
						}
					}
				}
			}
		}

		public bool HasLoweredCeiling(Polygon poly)
		{
			var polyIndex = GetPolygonIndex(poly);
			for (int i = 0; i < poly.LineIndexes.Length; i++)
			{
				if (poly.LineIndexes[i] > -1)
				{
					var line = _level.Lines[poly.LineIndexes[i]];
					if (line.ClockwisePolygonOwner > -1 && line.ClockwisePolygonOwner != polyIndex)
					{
						var p = _level.Polygons[line.ClockwisePolygonOwner];
						if (p.CeilingHeight > poly.CeilingHeight)
							return true;
					}
					if (line.CounterclockwisePolygonOwner > -1 && line.CounterclockwisePolygonOwner != polyIndex)
					{
						var p = _level.Polygons[line.CounterclockwisePolygonOwner];
						if (p.CeilingHeight > poly.CeilingHeight)
							return true;
					}
				}
			}
			return false;
		}

		public void AutoShade(short highlight, short shadow)
		{
			for (int side_index = 0; side_index < _level.Sides.Count; ++side_index)
			{
				Side side = _level.Sides[side_index];

				if (side.PrimaryLightsourceIndex > 20 || side.SecondaryLightsourceIndex > 20)
					continue;

				Line line = _level.Lines[side.LineIndex];
				double theta;
				if (line.ClockwisePolygonSideIndex == side_index)
				{
					theta = Angle(_level.Endpoints[line.EndpointIndexes[0]], _level.Endpoints[line.EndpointIndexes[1]]);
				}
				else
				{
					theta = Angle(_level.Endpoints[line.EndpointIndexes[1]], _level.Endpoints[line.EndpointIndexes[0]]);
				}

				var lightsource = (short)(shadow + (highlight - shadow) * ((Math.Cos(theta) + 1) / 2));
				side.PrimaryLightsourceIndex = lightsource;
				side.SecondaryLightsourceIndex = lightsource;
			}
		}

		public static double Angle(Point p0, Point p1)
		{
			int X = p1.X - p0.X;
			int Y = p1.Y - p0.Y;
			var theta = Math.Atan2(-Y, X);
			if (theta < 0)
			{
				return -theta;
			}

			return theta;
		}

		public short GetPolygonIndex(Polygon poly)
		{
			short index = -1;
			foreach (var p in _level.Polygons)
			{
				index++;
				if (p == poly)
					return index;
			}
			throw new Exception("Polygon not found");
		}
	}
}