using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using Weland;
using Gtk;
using System.Drawing;

namespace StairGenerator
{
	public class Plugin
	{
		public static bool Compatible()
		{
			return true;
		}

		public static string Name()
		{
			return "Stair Generator";
		}

		public static void GtkRun(Editor editor)
		{
			if (editor.Selection != null && editor.Selection.Line != -1)
			{
				new StairGenerator(editor);
			}
			else
			{
				MessageDialog md = new MessageDialog(null, DialogFlags.Modal, MessageType.Info, ButtonsType.Ok,
									   "You must select a line which will tell the plugin where the steps should be, and how long to make them.");
				md.Run();
				md.Destroy();
			}
		}
	}

	class StairGenerator : Window
	{

		Editor _editor;

		CheckButton chkFloorHeightProvided;
		CheckButton chkCeilingHeightProvided;
		Entry txtFloorOfEach;
		Entry txtCeilingOfEach;
		Entry txtFloorOfLast;
		Entry txtCeilingOfLast;

		Entry txtNumSteps;
		Entry txtWidthFirst;
		Entry txtWidthLast;
		Entry txtFloorOfFirst;
		Entry txtCeilingOfFirst;

		public StairGenerator(Editor editor) : base("Generate Stairs")
		{
			_editor = editor;

			// Horrible GTK code
			SetDefaultSize(350, 400);
			SetPosition(WindowPosition.Center);
			this.BorderWidth = 0;

			VBox box1 = new VBox(false, 0);

			Label lblSteps = new Label();
			lblSteps.Text = "Number of steps:";
			box1.PackStart(lblSteps, true, true, 0);
			lblSteps.Show();

			txtNumSteps = new Entry();
			box1.PackStart(txtNumSteps, true, true, 0);
			txtNumSteps.Show();

			Label blankLabel = new Label();
			box1.PackStart(blankLabel, true, true, 3);
			blankLabel.Show();

			Label lblWidthFirst = new Label();
			lblWidthFirst.Text = "Width of first step:";
			box1.PackStart(lblWidthFirst, true, true, 0);
			lblWidthFirst.Show();

			txtWidthFirst = new Entry();
			box1.PackStart(txtWidthFirst, true, true, 0);
			txtWidthFirst.Show();

			Label lblWidthLast = new Label();
			lblWidthLast.Text = "Width of last step:";
			box1.PackStart(lblWidthLast, true, true, 0);
			lblWidthLast.Show();

			txtWidthLast = new Entry();
			box1.PackStart(txtWidthLast, true, true, 0);
			txtWidthLast.Show();

			blankLabel = new Label();
			box1.PackStart(blankLabel, true, true, 3);
			blankLabel.Show();

			Label lblFloors = new Label();
			lblFloors.Text = "Floors";
			box1.PackStart(lblFloors, true, true, 0);
			lblFloors.Show();

			Label lblFloorOfFirst = new Label();
			lblFloorOfFirst.Text = "Height of first step:";
			box1.PackStart(lblFloorOfFirst, true, true, 0);
			lblFloorOfFirst.Show();

			txtFloorOfFirst = new Entry();
			box1.PackStart(txtFloorOfFirst, true, true, 0);
			txtFloorOfFirst.Show();

			chkFloorHeightProvided = new CheckButton("Specify height of each step");
			chkFloorHeightProvided.Clicked += floor_height_provided_event;
			box1.PackStart(chkFloorHeightProvided, true, true, 0);
			chkFloorHeightProvided.Show();

			Label lblFloorOfLast = new Label();
			lblFloorOfLast.Text = "Height of last step:";
			box1.PackStart(lblFloorOfLast, true, true, 0);
			lblFloorOfLast.Show();

			txtFloorOfLast = new Entry();
			box1.PackStart(txtFloorOfLast, true, true, 0);
			txtFloorOfLast.Show();

			Label lblFloorOfEach = new Label();
			lblFloorOfEach.Text = "Height of each step:";
			box1.PackStart(lblFloorOfEach, true, true, 0);
			lblFloorOfEach.Show();

			txtFloorOfEach = new Entry();
			txtFloorOfEach.IsEditable = false;
			txtFloorOfEach.ModifyBase(StateType.Normal, new Gdk.Color(150, 150, 150));
			box1.PackStart(txtFloorOfEach, true, true, 0);
			txtFloorOfEach.Show();

			blankLabel = new Label();
			box1.PackStart(blankLabel, true, true, 3);
			blankLabel.Show();

			Label lblCeilings = new Label();
			lblCeilings.Text = "Ceilings";
			box1.PackStart(lblCeilings, true, true, 0);
			lblCeilings.Show();

			Label lblCeilingOfFirst = new Label();
			lblCeilingOfFirst.Text = "Height of first step:";
			box1.PackStart(lblCeilingOfFirst, true, true, 0);
			lblCeilingOfFirst.Show();

			txtCeilingOfFirst = new Entry();
			box1.PackStart(txtCeilingOfFirst, true, true, 0);
			txtCeilingOfFirst.Show();

			chkCeilingHeightProvided = new CheckButton("Specify height of each step");
			chkCeilingHeightProvided.Clicked += ceiling_height_provided_event;
			box1.PackStart(chkCeilingHeightProvided, true, true, 0);
			chkCeilingHeightProvided.Show();

			Label lblCeilingOfLast = new Label();
			lblCeilingOfLast.Text = "Height of last step:";
			box1.PackStart(lblCeilingOfLast, true, true, 0);
			lblCeilingOfLast.Show();

			txtCeilingOfLast = new Entry();
			box1.PackStart(txtCeilingOfLast, true, true, 0);
			txtCeilingOfLast.Show();

			Label lblCeilingOfEach = new Label();
			lblCeilingOfEach.Text = "Height of each step:";
			box1.PackStart(lblCeilingOfEach, true, true, 0);
			lblCeilingOfEach.Show();

			txtCeilingOfEach = new Entry();
			txtCeilingOfEach.IsEditable = false;
			txtCeilingOfEach.ModifyBase(StateType.Normal, new Gdk.Color(150, 150, 150));
			box1.PackStart(txtCeilingOfEach, true, true, 0);
			txtCeilingOfEach.Show();

			blankLabel = new Label();
			box1.PackStart(blankLabel, true, true, 3);
			blankLabel.Show();

			Button button = new Button("Generate");
			button.Clicked += create_stairs_event;
			box1.PackStart(button, true, true, 0);
			button.Show();

			this.Add(box1);
			box1.Show();

			this.ShowAll();
		}

		void create_stairs_event(object obj, EventArgs args)
		{
			try
			{
				_editor.SetUndo();

				// Convert from WU to internal units.
				short numberOfSteps = Convert.ToInt16(txtNumSteps.Text);
				short widthOfFirst = (short)(Convert.ToDouble(txtWidthFirst.Text) * 1024);
				short widthOfLast = (short)(Convert.ToDouble(txtWidthLast.Text) * 1024);

				bool ceilingHeightOfEachProvided = chkCeilingHeightProvided.Active;
				short ceilingHeightOfFirst = (short)(Convert.ToDouble(txtCeilingOfFirst.Text) * 1024);
				short ceilingHeightOfLast;
				short ceilingHeightOfEach;
				if (ceilingHeightOfEachProvided)
				{
					ceilingHeightOfLast = 0;
					ceilingHeightOfEach = (short)(Convert.ToDouble(txtCeilingOfEach.Text) * 1024);
				}
				else
				{
					ceilingHeightOfLast = (short)(Convert.ToDouble(txtCeilingOfLast.Text) * 1024);
					ceilingHeightOfEach = 0;
				}

				bool floorHeightOfEachProvided = chkFloorHeightProvided.Active;
				short floorHeightOfFirst = (short)(Convert.ToDouble(txtFloorOfFirst.Text) * 1024);
				short floorHeightOfLast;
				short floorHeightOfEach;
				if (floorHeightOfEachProvided)
				{
					floorHeightOfLast = 0;
					floorHeightOfEach = (short)(Convert.ToDouble(txtFloorOfEach.Text) * 1024);
				}
				else
				{
					floorHeightOfLast = (short)(Convert.ToDouble(txtFloorOfLast.Text) * 1024);
					floorHeightOfEach = 0;
				}

				CreateStairs(numberOfSteps, widthOfFirst, widthOfLast, ceilingHeightOfEachProvided,
					ceilingHeightOfFirst, ceilingHeightOfLast, ceilingHeightOfEach,
					floorHeightOfEachProvided, floorHeightOfFirst, floorHeightOfLast, floorHeightOfEach);
				this.Destroy();
			}
			catch
			{ // Having a prahblum?
				MessageDialog md = new MessageDialog(null, DialogFlags.Modal, MessageType.Info, ButtonsType.Ok,
									   "Prahblum detected!\n\nMake sure you've entered sane numbers. Do it now.");
				md.Run();
				md.Destroy();
				_editor.Undo();
			}
		}

		void floor_height_provided_event(object obj, EventArgs args)
		{
			if (chkFloorHeightProvided.Active)
			{
				txtFloorOfEach.IsEditable = true;
				txtFloorOfEach.ModifyBase(StateType.Normal, new Gdk.Color(255, 255, 255));

				txtFloorOfLast.IsEditable = false;
				txtFloorOfLast.Text = "";
				txtFloorOfLast.ModifyBase(StateType.Normal, new Gdk.Color(150, 150, 150));
			}
			else
			{
				txtFloorOfLast.IsEditable = true;
				txtFloorOfLast.ModifyBase(StateType.Normal, new Gdk.Color(255, 255, 255));

				txtFloorOfEach.IsEditable = false;
				txtFloorOfEach.Text = "";
				txtFloorOfEach.ModifyBase(StateType.Normal, new Gdk.Color(150, 150, 150));
			}
		}

		void ceiling_height_provided_event(object obj, EventArgs args)
		{
			if (chkCeilingHeightProvided.Active)
			{
				txtCeilingOfEach.IsEditable = true;
				txtCeilingOfEach.ModifyBase(StateType.Normal, new Gdk.Color(255, 255, 255));

				txtCeilingOfLast.IsEditable = false;
				txtCeilingOfLast.Text = "";
				txtCeilingOfLast.ModifyBase(StateType.Normal, new Gdk.Color(150, 150, 150));
			}
			else
			{
				txtCeilingOfLast.IsEditable = true;
				txtCeilingOfLast.ModifyBase(StateType.Normal, new Gdk.Color(255, 255, 255));

				txtCeilingOfEach.IsEditable = false;
				txtCeilingOfEach.Text = "";
				txtCeilingOfEach.ModifyBase(StateType.Normal, new Gdk.Color(150, 150, 150));
			}
		}

		private void CreateStairs(
			short numberOfSteps,
			short widthOfFirst,
			short widthOfLast,
			bool ceilingHeightOfEachProvided,
			short ceilingHeightOfFirst,
			short ceilingHeightOfLast,
			short ceilingHeightOfEach,
			bool floorHeightOfEachProvided,
			short floorHeightOfFirst,
			short floorHeightOfLast,
			short floorHeightOfEach)
		{
			Level level = _editor.Level;

			short xStart;
			short xEnd;
			short yStart;
			short yEnd;
			short xStart2;
			short xEnd2;
			short yStart2;
			short yEnd2;
			GetInfoAboutLines(out xStart, out xEnd, out yStart, out yEnd,
				out xStart2, out xEnd2, out yStart2, out yEnd2,
				widthOfFirst, widthOfLast);

			// Delete the line
			_editor.DeleteSelected();

			short numberOfPoints = (short)(numberOfSteps + 1);
			int xDiff = Math.Abs(xStart - xEnd);
			short xChange = (short)(xDiff / numberOfSteps);
			int yDiff = Math.Abs(yStart - yEnd);
			short yChange = (short)(yDiff / numberOfSteps);

			List<short> firstPoints = new List<short>();
			List<short> firstLines = new List<short>();
			List<short> acrossLines = new List<short>();
			List<short> polygons = new List<short>();

			// Create the correct number of points along the original line, and create the lines connecting them.
			CreateSidePointsAndLines(firstPoints, firstLines, numberOfPoints,
				xStart, xEnd, yStart, yEnd, xChange, yChange);

			xDiff = Math.Abs(xStart2 - xEnd2);
			xChange = (short)(xDiff / numberOfSteps);
			yDiff = Math.Abs(yStart2 - yEnd2);
			yChange = (short)(yDiff / numberOfSteps);

			List<short> secondPoints = new List<short>();
			List<short> secondLines = new List<short>();

			CreateSidePointsAndLines(secondPoints, secondLines, numberOfPoints,
				xStart2, xEnd2, yStart2, yEnd2, xChange, yChange);

			// Connect lines across
			for (short i = 0; i < numberOfSteps + 1; i++)
			{
				acrossLines.Add(level.NewLine(firstPoints[i], secondPoints[i]));
			}

			// Fill the polygons
			for (short i = 0; i < numberOfSteps; i++)
			{
				List<Weland.Point> points = new List<Weland.Point>();
				points.Add(level.Endpoints[firstPoints[i]]);
				points.Add(level.Endpoints[firstPoints[i + 1]]);
				points.Add(level.Endpoints[secondPoints[i]]);
				points.Add(level.Endpoints[secondPoints[i + 1]]);
				Weland.Point center = GetPolygonCenter(points);
				level.FillPolygon(center.X, center.Y);
				polygons.Add(level.GetEnclosingPolygon(new Weland.Point(center.X, center.Y)));
			}

			// Calculate ceiling/floor height difference between each step
			short ceilingHeightDiff;
			short floorHeightDiff;
			if (ceilingHeightOfEachProvided)
				ceilingHeightDiff = ceilingHeightOfEach;
			else
				ceilingHeightDiff = (short)((ceilingHeightOfLast - ceilingHeightOfFirst) / (numberOfSteps - 1));

			if (floorHeightOfEachProvided)
				floorHeightDiff = floorHeightOfEach;
			else
				floorHeightDiff = (short)((floorHeightOfLast - floorHeightOfFirst) / (numberOfSteps - 1));

			// Set floor/ceiling heights
			short currentFloorHeight = floorHeightOfFirst;
			short currentCeilingHeight = ceilingHeightOfFirst;
			foreach (short polygonIndex in polygons)
			{
				Polygon polygon = level.Polygons[polygonIndex];
				polygon.FloorHeight = currentFloorHeight;
				polygon.CeilingHeight = currentCeilingHeight;

				currentFloorHeight += floorHeightDiff;
				currentCeilingHeight += ceilingHeightDiff;
			}

			// Try to correct any rounding errors for the last step
			if (!ceilingHeightOfEachProvided)
			{
				Polygon polygon = level.Polygons[polygons[polygons.Count - 1]];
				polygon.CeilingHeight = ceilingHeightOfLast;
			}
			if (!floorHeightOfEachProvided)
			{
				Polygon polygon = level.Polygons[polygons[polygons.Count - 1]];
				polygon.FloorHeight = floorHeightOfLast;
			}
		}

		private void CreateSidePointsAndLines(List<short> points,
											   List<short> lines,
											   short numberOfPoints,
											   short xStart,
											   short xEnd,
											   short yStart,
											   short yEnd,
											   short xChange,
											   short yChange)
		{
			// Create points for one side of the stairs
			for (short i = 0; i < numberOfPoints; i++)
			{
				short newPoint;

				short newX;
				short newY;
				if (xStart < xEnd)
					newX = (short)(xStart + (i * xChange));
				else
					newX = (short)(xStart - (i * xChange));

				if (yStart < yEnd)
					newY = (short)(yStart + (i * yChange));
				else
					newY = (short)(yStart - (i * yChange));

				newPoint = _editor.Level.NewPoint(newX, newY);
				points.Add(newPoint);
			}

			// Connect the points along the side of the stairs
			for (short i = 0; i < numberOfPoints; i++)
			{
				if (i + 1 < points.Count)
				{
					short newLine = _editor.Level.NewLine(points[i], points[i + 1]);
					lines.Add(newLine);
				}
			}
		}

		private void GetInfoAboutLines(out short xStart1,
										out short xEnd1,
										out short yStart1,
										out short yEnd1,
										out short xStart2,
										out short xEnd2,
										out short yStart2,
										out short yEnd2,
										short widthOfFirst,
										short widthOfLast)
		{
			Line selectedLine = _editor.Level.Lines[_editor.Selection.Line];

			short xStart, xEnd, yStart, yEnd;

			xStart = _editor.Level.Endpoints[selectedLine.EndpointIndexes[0]].X;
			xEnd = _editor.Level.Endpoints[selectedLine.EndpointIndexes[1]].X;
			yStart = _editor.Level.Endpoints[selectedLine.EndpointIndexes[0]].Y;
			yEnd = _editor.Level.Endpoints[selectedLine.EndpointIndexes[1]].Y;

			double slope;
			if ((double)(xStart - xEnd) == 0)
				slope = double.MaxValue;
			else
				slope = (double)((yStart - yEnd)) / (double)((xStart - xEnd));

			slope = 1 / slope * -1.0;
			double atan = Math.Atan(slope);

			xStart1 = (short)(xStart - (Math.Cos(atan) * (widthOfFirst / 2)));
			xEnd1 = (short)(xEnd - (Math.Cos(atan) * (widthOfLast / 2)));
			yStart1 = (short)(yStart - (Math.Sin(atan) * (widthOfFirst / 2)));
			yEnd1 = (short)(yEnd - (Math.Sin(atan) * (widthOfLast / 2)));

			xStart2 = (short)(xStart + (Math.Cos(atan) * (widthOfFirst / 2)));
			xEnd2 = (short)(xEnd + (Math.Cos(atan) * (widthOfLast / 2)));
			yStart2 = (short)(yStart + (Math.Sin(atan) * (widthOfFirst / 2)));
			yEnd2 = (short)(yEnd + (Math.Sin(atan) * (widthOfLast / 2)));
		}

		private Weland.Point GetPolygonCenter(List<Weland.Point> points)
		{
			double x = 0;
			double y = 0;
			for (int i = 0; i < points.Count; ++i)
			{
				x += points[i].X;
				y += points[i].Y;
			}

			return new Weland.Point((short)(x / points.Count), (short)(y / points.Count));
		}
	}
}